/*
 * iflib_arith.c - Arithmetic conversions
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 EPFL-ICA, Network Robots
 */


#include "data.h"
#include "op.h"
#include "iflib.h"


/*
 * General naming conventions in this file:
 *
 * X = general expression
 * C = constant
 * rel = relational operator
 */

/*
 * @@@ several of these optimizations should be either done in op.c, or the
 * optimizations in op.c should move here.
 */


/* --------------------------- Helper functions ---------------------------- */


static int log_2(uint32_t v)
{
    int ld = 0;

    if (!v) return -1;
    while (!(v & 1)) {
	ld++;
	v >>= 1;
    }
    return v == 1 ? ld : -1;
}


/* ------------------------------ *, /, and % ------------------------------ */


/*
 * Convert:
 *
 *  0 * X  ->  0
 *  X * 0  ->  0
 *  1 * X  ->  X
 *  X * 1  ->  X
 *  (2^n) * X  ->  X << n
 *  X * (2^n)  ->  X << n
 *
 *  X / 1      ->  X
 *  0 / X      ->  0
 *  X / X      ->  1
 *  X / (2^n)  ->  X >> n
 *
 *  X % 1      ->  0
 *  X % (2^n)  ->  X & (2^n)-1
 *  X % X      ->  0
 */


static int mul_opt(DATA *d,DATA x,DATA *c,uint8_t value)
{
    int ld;

    if (!value) {
	data_destroy(*d);
	*d = data_unum(0);
	return 1;
    }
    ld = log_2(value);
    if (ld == -1) return 0;
    data_destroy(c);
    data_destroy_1(d);
    if (!ld) {
	*d = x;
	return 1;
    }
    *d = op_binary(&op_shift_left,x,data_unum(ld));
    return 1;
}


static int div_l_opt(DATA *d,DATA x,DATA *c,uint8_t value)
{
    if (value) return 0;
    data_destroy(*d);
    *d = data_unum(0);
    return 1;
}


static int div_r_opt(DATA *d,DATA x,DATA *c,uint8_t value)
{
    int ld;

    ld = log_2(value);
    if (ld == -1) return 0;
    data_destroy(c);
    data_destroy_1(d);
    if (!ld) {
	*d = x;
	return 1;
    }
    *d = op_binary(&op_shift_right,x,data_unum(ld));
    return 1;
}


static int mod_opt(DATA *d,DATA x,DATA *c,uint8_t value)
{
    int ld;

    if (value == 1) {
	data_destroy(*d);
	*d = data_unum(0);
	return 1;
    }
    ld = log_2(value);
    if (ld == -1) return 0;
    d->op->dsc = &op_and;
    c->u.unum--;
    return 1;
}


static void mul_div_mod(DATA *d)
{
    if (!d->op) return;
    mul_div_mod(&d->op->a);
    mul_div_mod(&d->op->b);
    mul_div_mod(&d->op->c);
    if (d->op->dsc == &op_mult) {
	if (!d->op->a.op && d->op->a.type == dt_unum) {
	    if (mul_opt(d,d->op->b,&d->op->a,d->op->a.u.unum)) mul_div_mod(d);
	}
	else if (!d->op->b.op && d->op->b.type == dt_unum) {
	    if (mul_opt(d,d->op->a,&d->op->b,d->op->b.u.unum)) mul_div_mod(d);
	}
    }
    else if (d->op->dsc == &op_div) {
	if (!d->op->a.op && d->op->a.type == dt_unum) {
	    if (div_l_opt(d,d->op->b,&d->op->a,d->op->a.u.unum))
		mul_div_mod(d);
	}
	else if (!d->op->b.op && d->op->b.type == dt_unum) {
	    if (div_r_opt(d,d->op->a,&d->op->b,d->op->b.u.unum))
		mul_div_mod(d);
	}
	else if (expr_equal(d->op->a,d->op->b)) {
		    data_destroy(*d);
		    *d = data_unum(1);
	    }
    }
    else if (d->op->dsc == &op_mod) {
	if (!d->op->b.op && d->op->b.type == dt_unum) {
	    if (mod_opt(d,d->op->a,&d->op->b,d->op->b.u.unum)) mul_div_mod(d);
	}
	else if (expr_equal(d->op->a,d->op->b)) {
		data_destroy(*d);
		*d = data_unum(0);
	    }
    }
}

/* ----------------- Relational operators (move constants) ----------------- */


/*
 * Convert:
 *
 * C1 rel C2  ->  0  if !(C1 rel C2)
 * C1 rel C2  ->  1  if C1 rel C2
 *
 * C rel X  ->  X inv rel C  (exchange < and >, <= and >=)
 *
 * X+C1 rel C2  ->  X rel C1-C2
 * C1+X rel C2  ->  X rel C1-C2
 *
 * X-C1 rel C2  ->  X rel C1+C2
 * C1-X rel C2  ->  X rel C1-C2
 */


static void rel_plus(DATA *d,DATA *a,DATA *b)
{
    DATA *a_c,x;

    if (!a->op->a.op && a->op->a.type == dt_unum) {
	a_c = &a->op->a;
	x = a->op->b;
    }
    else if (!a->op->b.op && a->op->b.type == dt_unum) {
	x = a->op->a;
	a_c = &a->op->b;
    }
    else return;
    b->u.unum -= a_c->u.unum;
    data_destroy_1(*a_c);
    data_destroy_1(*a);
    *a = x;
}


static void rel_minus(DATA *d,DATA *a,DATA *b)
{
    DATA *a_c,x;

    if (!a->op->a.op && a->op->a.type == dt_unum) {
	a_c = &a->op->a;
	x = a->op->b;
	b->u.unum = a_c->u.unum-b->u.unum;
    }
    else if (!a->op->b.op && a->op->b.type == dt_unum) {
	x = a->op->a;
	a_c = &a->op->b;
	b->u.unum += a_c->u.unum;
    }
    else return;
    data_destroy_1(*a_c);
    data_destroy_1(*a);
    *a = x;
}


static void rel_math(DATA *d)
{
    if (!d->op) return;
    if (d->op->dsc == &op_eq || d->op->dsc == &op_ne || d->op->dsc == &op_gt ||
      d->op->dsc == &op_ge || d->op->dsc == &op_lt || d->op->dsc == &op_le) {
	if (!d->op->a.op && !d->op->b.op && d->op->a.type == dt_unum &&
	  d->op->b.type == dt_unum) {
	    DATA tmp = op_binary(d->op->dsc,d->op->a,d->op->b);

	    data_destroy(*d);
	    *d = tmp;
	    return;
	}
	if (!d->op->a.op && d->op->a.type == dt_unum && d->op->b.op) {
	    DATA tmp;

	    tmp = d->op->a;
	    d->op->a = d->op->b;
	    d->op->b = tmp;
	    if (d->op->dsc == &op_gt) d->op->dsc = &op_lt;
	    else if (d->op->dsc == &op_lt) d->op->dsc = &op_gt;
	    else if (d->op->dsc == &op_ge) d->op->dsc = &op_le;
	    else if (d->op->dsc == &op_le) d->op->dsc = &op_ge;
	}
	if (d->op->a.op && d->op->a.op->dsc == &op_plus && !d->op->b.op &&
	  d->op->b.type == dt_unum) {
	    rel_plus(d,&d->op->a,&d->op->b);
	    return;
	}
	if (d->op->a.op && d->op->a.op->dsc == &op_minus && !d->op->b.op &&
	  d->op->b.type == dt_unum) {
	    rel_minus(d,&d->op->a,&d->op->b);
	    return;
	}
    }
    rel_math(&d->op->a);
    rel_math(&d->op->b);
    rel_math(&d->op->c);
}


/* ------------ Relational operators (turn them into equality) ------------- */


/*
 * Convert:
 *
 * X < 2^n  ->  X & ~(2^n-1) == 0
 * X <= 2^n-1  ->  X & ~(2^n-1) == 0
 * X > 2^n-1  ->  X & ~(2^n-1) != 0	requires handling of !
 * X >= 2^n  ->  X & ~(2^n-1) != 0	requires handling of !
 */


static void rel_to_eq(DATA *d)
{
    if (!d->op) return;
    rel_to_eq(&d->op->a);
    rel_to_eq(&d->op->b);
    rel_to_eq(&d->op->c);
    if (d->op->a.op && !d->op->b.op && d->op->b.type == dt_unum) {
	uint32_t mask;
	int ld;

	if (d->op->dsc == &op_lt) {
	    ld = log_2(d->op->b.u.unum);
	    if (ld == -1) return;
	    mask = d->op->b.u.unum-1;
	    d->op->dsc = &op_eq;
	}
	else if (d->op->dsc == &op_le) {
	    ld = log_2(d->op->b.u.unum+1);
	    if (ld == -1 && (d->op->b.u.unum+1)) return; /* roll-over */
	    mask = d->op->b.u.unum;
	    d->op->dsc = &op_eq;
	}
	else if (d->op->dsc == &op_gt) {
	    ld = log_2(d->op->b.u.unum+1);
	    if (ld == -1 && (d->op->b.u.unum+1)) return; /* roll-over */
	    mask = d->op->b.u.unum;
	    d->op->dsc = &op_ne;
	}
	else if (d->op->dsc == &op_ge) {
	    ld = log_2(d->op->b.u.unum);
	    if (ld == -1) return;
	    mask = d->op->b.u.unum-1;
	    d->op->dsc = &op_ne;
	}
	else return;
	d->op->a = op_binary(&op_and,d->op->a,data_unum(~mask));
	d->op->b.u.unum = 0;
    }
}


/* ----------------- Eliminate expressions involving |, ^ ------------------ */


/*
 * Convert:
 *
 * X == X	 ->  1
 * X ^ C1 == C2  ->  X == C1 ^ C2
 * X | C1 == C2  ->  0				if (C1 & C2) != C1
 * X | C1 == C2  ->  X & ~C1 == (C2 & ~C1)	else
 *
 * ... also for C1 op X
 */


static void rel_eq_or(DATA *d)
{
    DATA x,c;

    if (!d->op) return;
    rel_eq_or(&d->op->a);
    rel_eq_or(&d->op->b);
    rel_eq_or(&d->op->c);
    if (d->op->dsc == &op_eq && expr_equal(d->op->a,d->op->b)) {
	data_destroy(*d);
	*d = data_unum(1);
	return;
    }
    if (d->op->dsc != &op_eq || d->op->b.op || d->op->b.type != dt_unum)
	return;
    if (!d->op->a.op || (d->op->a.op->dsc != &op_or &&
      d->op->a.op->dsc != &op_xor)) return;
    if (!d->op->a.op->a.op && d->op->a.op->a.type == dt_unum) {
	c = d->op->a.op->a;
	x = d->op->a.op->b;
    }
    else if (!d->op->a.op->b.op && d->op->a.op->b.type == dt_unum) {
	x = d->op->a.op->a;
	c = d->op->a.op->b;
    }
    else return;
    if (d->op->a.op->dsc == &op_xor) {
	d->op->b.u.unum ^= c.u.unum;
	data_destroy_1(d->op->a);
	data_destroy_1(c);
	d->op->a = x;
	return;
    }
    if ((c.u.unum & d->op->b.u.unum) != c.u.unum) {
	data_destroy(*d);
	*d = data_unum(0);
	return;
    }
    d->op->a.op->dsc = &op_and;
    d->op->a.op->a = x;
    d->op->a.op->b = data_unum(~c.u.unum);
    d->op->b.u.unum &= ~c.u.unum;
}


/* ------------------------------------------------------------------------- */


void iflib_arith(DATA *d)
{
    mul_div_mod(d);
    debug_expr("AFTER mul_div_mod",*d);
    rel_math(d);
    debug_expr("AFTER rel_math",*d);
    rel_to_eq(d);
    debug_expr("AFTER rel_to_eq",*d);
    rel_eq_or(d);
    debug_expr("AFTER rel_eq_or",*d);
}
